#include "videoshow.h"

#define MODE_PLAY 1
#define MODE_PAUSE 2
#define MODE_STEP 3
#define MODE_NEXT 4

SDL_Surface *SDLVideo = NULL;

static int SDL_BPP = 16;
static int VideoStack = 1;
static int WriteFrame = 0;
int Mode = MODE_PLAY;
int VideoCut = 0;
int Offset = 0;
double UMask = 0.0;
int AMean = 0;

extern int DetectBrokenFrames;

int DrawFrame = 1;
int SeekColour = 0;

char * GainComp = NULL;

unsigned long FrameOffsets[10000];
int FrameCount = 0;

// 6 channels, 1..6 0 is the default
#define MAX_CHANNEL 7

// Map input buffer value to screen value, keep array of mappings so each
// wheel channel can be handles separately
struct gstruct {
	unsigned int *map;
	double gamma;
	double gain;
	int bpp;
	};

struct gstruct Tables[MAX_CHANNEL];
int Channel = 0;

static int FrameChannel(char *fname);
static void Sharpen(unsigned int *buffer, int width, int height);
static void Smooth(unsigned int *buffer, int width, int height);

static int create_tables(void);
static int create_table_16bpp_565(int tbl);
static int create_table_16bpp_888(int tbl);

int isBrokenFrame(struct Image *img);

static int
create_tables()
	{
	int i;

	for(i=0; i<MAX_CHANNEL; ++i) {
	   if (SDL_BPP == 16) create_table_16bpp_565(i);
	   if (SDL_BPP == 32) create_table_16bpp_888(i);
	   }

	return 1;
	}

/*
 * create a lookup table that maps a value (multiplied by gain) to 565
 */
static int
create_table_16bpp_565(int tbl)
        {
        int i;
	double gain = Tables[tbl].gain;
	double gamma = Tables[tbl].gamma;
	unsigned int *map;

	if (Tables[tbl].map == NULL) 
	   Tables[tbl].map = (unsigned int *)Malloc(65536 * 4);

	map = Tables[tbl].map;

        for(i=0; i<65536; ++i) {
	   int n = i * gain;
	   if (n>65535) n = 65535;
           unsigned int v = (unsigned int) (pow((double)n / 65535.0, 1.0/gamma) * 65535);
           if (v>65535) v=65535;

	   // convert to 5-6-5 format
           map[i] = (v&0xf800) | ((v&0xfc00)>>5) | (v>>11);
           }

        return(1);
        }

static int
create_table_16bpp_888(int tbl)
        {
        int i;
	double gain = Tables[tbl].gain;
	double gamma = Tables[tbl].gamma;
	unsigned int *map;

	if (Tables[tbl].map == NULL) 
	   Tables[tbl].map = (unsigned int *)Malloc(65536 * 4);

	map = Tables[tbl].map;

        for(i=0; i<65536; ++i) {
	   int n = i * gain;
	   if (n>65535) n = 65535;
           unsigned int v = (unsigned int) (pow((double)n / 65535.0, 1.0/gamma) * 65535);
           if (v>65535) v=65535;

	   v >>= 8;

	   // convert to 5-6-5 format
           map[i] = v | (v<<8) | (v<<16);
           }

        return(1);
        }

SDL_Overlay *
init_display(int width, int height, int bpp)
	{
	const SDL_VideoInfo  *sdlvideoinfo;
	SDL_Overlay *sdloverlay;
	SDL_Rect **modes;
	unsigned int sdlflags;
	int sdlbpp;
	xvinfo_t xvinfo;

	GetXvInfo(&xvinfo);

	if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_NOPARACHUTE) == -1) {
	   FatalError("Could not init SDL\n");
	   exit(1);
	   }

	sdlvideoinfo = SDL_GetVideoInfo();

	sdlflags = SDL_ANYFORMAT | SDL_RESIZABLE;

  	if ((xvinfo.max_width!=-1)&&(xvinfo.max_height!=-1)) {
    	   // if the XV area is too small, we use software acceleration
    	   if ((xvinfo.max_width < width)|| (xvinfo.max_height < height)) {
      	   	//Message("SDL: Using SW surface");
      	   	sdlflags|= SDL_SWSURFACE;
      	   	sdlflags&= ~SDL_HWSURFACE;
      	   	sdlflags&= ~SDL_HWACCEL;
    	   	}
    	   else {
      	   	//Message("SDL: Using HW surface");
      	   	sdlflags|= SDL_HWSURFACE;
      	   	sdlflags|= SDL_HWACCEL;
      	   	sdlflags&= ~SDL_SWSURFACE;
    	   	}
  	   }
  	else {
    	   // try HW accel and pray...
      	   //Message("SDL: Using any surface");
    	   sdlflags|= SDL_SWSURFACE;
    	   sdlflags&= ~SDL_HWSURFACE;
    	   sdlflags&= ~SDL_HWACCEL;
  	   }

	modes = SDL_ListModes(NULL,sdlflags);
  	if (modes!=(SDL_Rect**)-1) {
    	   // not all resolutions are OK for this video card. 
	   // For safety we switch to software accel
    	   Message("SDL: No SDL mode available with hardware accel. Trying without HWSURFACE");
    	   sdlflags&= ~SDL_HWSURFACE;
    	   sdlflags&= ~SDL_HWACCEL;
    	   modes=SDL_ListModes(NULL,sdlflags);
    	   if (modes!=(SDL_Rect**)-1) {
      	      fprintf(stderr,"SDL: Still no modes available. Can't start SDL!\n");
      	      SDL_Quit();
      	      return(NULL);
    	      }
	   }
	
	sdlbpp = SDL_VideoModeOK(width, height, bpp, sdlflags);

	if (sdlbpp != bpp) {
	   fprintf(stderr,"SDL: Requested %d but SDL says %d is nearest bpp\n",
		bpp,sdlbpp);
	   SDL_Quit();
	   return(NULL);
	   }

	SDLVideo = SDL_SetVideoMode(width, height, bpp, sdlflags);

 	if (SDLVideo == NULL) {
    	   FatalError(SDL_GetError());
    	   return(NULL);
  	   }

  	if (SDL_SetColorKey( SDLVideo, SDL_SRCCOLORKEY, 0x0) < 0 ) {
    	   Message(SDL_GetError());
	   return(NULL);
  	   }

	SDL_ShowCursor(1);

	SDL_WM_SetCaption("showvideo","showvideo");

	 sdloverlay = SDL_CreateYUVOverlay(width, height, 
			SDL_YUY2_OVERLAY, SDLVideo);
  	if (sdloverlay==NULL) {
    	   FatalError(SDL_GetError());
    	   return(NULL);
  	   }
	
	SDL_EnableKeyRepeat(50,100);

	create_tables();
	return(sdloverlay);
	}

/* The following function is a strip-down version of the program 'xvinfo' of the XFree86 project. */
void
GetXvInfo(xvinfo_t *xvinfo) {

#ifdef HAVE_XV

  unsigned int ver, rev, eventB, reqB, errorB;
  int i, j, n;

  xvinfo->max_height=-1;
  xvinfo->max_width=-1;

  if(!(xvinfo->dpy = XOpenDisplay(NULL))) {
    FatalError("Unable to open display");
    return;
  }

  if((Success != XvQueryExtension(xvinfo->dpy, &ver, &rev, &reqB, &eventB, &errorB))) {
    //Message("xvinfo: No X-Video Extension");
    return;
  }

  i=0; // 0: we use only the first screen
  XvQueryAdaptors(xvinfo->dpy, RootWindow(xvinfo->dpy, i), &xvinfo->nadaptors, &xvinfo->ainfo);
  if( !xvinfo->nadaptors || !xvinfo->ainfo ){
    Message( "xvinfo: No adpators present\n" );
    return;
  }
  j=0; // 0: we use only the first adaptor
  XvQueryEncodings(xvinfo->dpy, xvinfo->ainfo[j].base_id, &xvinfo->nencode, &xvinfo->encodings);

  if(xvinfo->encodings && xvinfo->nencode) {
    xvinfo->ImageEncodings = 0;

    for(n = 0; n < xvinfo->nencode; n++) {
      if(!strcmp(xvinfo->encodings[n].name, "XV_IMAGE"))
        xvinfo->ImageEncodings++;
    }

    if(xvinfo->ImageEncodings && (xvinfo->ainfo[j].type & XvImageMask)) {
      //char imageName[5] = {0, 0, 0, 0, 0};

      for(n = 0; n < xvinfo->nencode; n++) {
        if(!strcmp(xvinfo->encodings[n].name, "XV_IMAGE")) {
          //fprintf(stdout, "maximum XvImage size: %li x %li\n", xvinfo->encodings[n].width, xvinfo->encodings[n].height);
          xvinfo->max_height=(int)xvinfo->encodings[n].height;
          xvinfo->max_width=(int)xvinfo->encodings[n].width;
          break;
        }
      }
      /*
      formats = XvListImageFormats(dpy, ainfo[j].base_id, &numImages);
      fprintf(stdout, "    Number of image formats: %i\n", numImages);

      for(n = 0; n < numImages; n++) {
        memcpy(imageName, &(formats[n].id), 4);
        fprintf(stdout, "      id: 0x%x", formats[n].id);
        if(isprint(imageName[0]) && isprint(imageName[1]) && isprint(imageName[2]) && isprint(imageName[3])) {
          fprintf(stdout, " (%s)\n", imageName);
        } else {
          fprintf(stdout, "\n");
        }
        fprintf(stdout, "        bits per pixel: %i\n", formats[n].bits_per_pixel);
        fprintf(stdout, "        number of planes: %i\n", formats[n].num_planes);
        fprintf(stdout, "        type: %s (%s)\n",
                (formats[n].type == XvRGB) ? "RGB" : "YUV",
                (formats[n].format == XvPacked) ? "packed" : "planar");
        if(formats[n].type == XvRGB) {
          fprintf(stdout, "        depth: %i\n", formats[n].depth);
          fprintf(stdout, "        red, green, blue masks: 0x%x, 0x%x, 0x%x\n",
                  formats[n].red_mask,
                  formats[n].green_mask,
                  formats[n].blue_mask);
        }
      }
      if(formats) XFree(formats);
      */
    }

    XvFreeEncodingInfo(xvinfo->encodings);
  }

  XvFreeAdaptorInfo(xvinfo->ainfo);

#else
  xvinfo->max_height=-1;
  xvinfo->max_width=-1;
#endif

  //fprintf(stderr,"%d %d\n", xvinfo->max_height, xvinfo->max_width);
}

void
shutdown_display()
	{
	SDL_Quit();
	}

void
Message(char *str)
	{
	fprintf(stderr,"%s\n",str);
	}

void
FatalError(char *str)
	{
	Message(str);
	exit(1);
	}

static void
VKeyPress(int key, int mod)
	{
	switch(key) {
	   case SDLK_ESCAPE:
	   case SDLK_q:
		SDL_Quit();
		exit(1);
		break;
	   case SDLK_n:
		Mode = MODE_NEXT;
		break;
	   case SDLK_SPACE:
		if (Mode == MODE_PLAY) Mode = MODE_PAUSE;
		else Mode = MODE_PLAY;
		break;
	   case SDLK_l:
	   case SDLK_RIGHT:
		Mode = MODE_STEP;
		break;
	   case SDLK_h:
	   case SDLK_LEFT:
		if (FrameCount>1) {
		   Mode = MODE_STEP;
		   FrameCount-=2;
		   }
		break;
	   case SDLK_2:
		DrawFrame=0;
		SeekColour = 2;
		break;
	   case SDLK_3:
		DrawFrame=0;
		SeekColour = 3;
		break;
	   case SDLK_4:
		DrawFrame=0;
		SeekColour = 4;
		break;
	   case SDLK_w:
		WriteFrame = 1;
		break;
	   }
	}

static void
VKeyRelease(int key, int mod)
	{
	}

int FCount=0;

void
SDLDisplay(struct Image *img)
	{
	unsigned short *data = (unsigned short *)img->data;
	unsigned short *p;
	int x,y,x1,y1,x2,y2;
	int i,o,width = img->width;
	int height = img->height;
	int npixels = width * height;
	unsigned int RShift,GShift,BShift,val, *bptr, *map;
	const SDL_PixelFormat *f = SDLVideo->format;
	static unsigned int *Buffer = NULL;
	static unsigned int *Buffer2 = NULL;
	static int Npixels = -1;
	int bpp = f->BytesPerPixel;

	x1 = img->cutout.x1;
	y1 = img->cutout.y1;
	x2 = img->cutout.x2;
	y2 = img->cutout.y2;

	if (npixels != Npixels) {
	   if (Buffer) free(Buffer);
	   if (Buffer2) free(Buffer2);
	   Buffer = (unsigned int *)ZeroMalloc(npixels * 4);
	   Buffer2 = (unsigned int *)ZeroMalloc(npixels * 4);
	   Npixels = npixels;
	   }

	Channel = FrameChannel(img->src_fname);
	if (Tables[Channel].map == NULL) create_tables();
	map = Tables[Channel].map;

	if (VideoStack<=1) {
	   for(y=y1; y<=y2; ++y) {
		o = y * width + x1;
		for(x=x1; x<=x2; ++x,++o) Buffer[o] = data[o];
	   	}
	   }
	else {
	   // intemediate buffer
	   data = (unsigned short *)img->data;
	   for(y=y1; y<=y2; ++y) {
		register int o = y * width + x1;
		for(x=x1; x<=x2; ++x,++o) {
		   register unsigned int b = Buffer[o];
		   b += data[o];
		   Buffer[o] = b/2;
		   }
		}
	   }
	   
	//memcpy(Buffer2,Buffer,npixels*4);
	SDL_LockSurface(SDLVideo);

	for(y=y1; y<=y2; ++y) {
	   register unsigned int *bptr = Buffer + y * width + x1;
    	   register unsigned short *ps = (unsigned short *)((Uint8 *)SDLVideo->pixels + y * SDLVideo->pitch + x1 * bpp);
    	   register unsigned int *pi = (unsigned int *)((Uint8 *)SDLVideo->pixels + y * SDLVideo->pitch + x1 * bpp);

	   switch(SDL_BPP) {
		case 16: for(x=x1; x<=x2; ++x) *(ps++) = (unsigned short)map[*(bptr++)];
			 break;
	   	case 32: for(x=x1; x<=x2; ++x) *(pi++) = map[*(bptr++)];
		 	 break;
		}
	   }

	// Also copy the last 20 rows as it may contain text
	for(y=height-20; 0 && y<height; ++y) {
	   register unsigned short *bptr = data + y * width + x1;
    	   register unsigned short *ps = (unsigned short *)((Uint8 *)SDLVideo->pixels + y * SDLVideo->pitch + x1 * bpp);
    	   register unsigned int *pi = (unsigned int *)((Uint8 *)SDLVideo->pixels + y * SDLVideo->pitch + x1 * bpp);

	   switch(SDL_BPP) {
		case 16: for(x=0; x<width; ++x) *(ps++) = (unsigned short)map[*(bptr++)];
			 break;
	   	case 32: for(x=0; x<width; ++x) *(pi++) = map[*(bptr++)];
		 	 break;
		}
	   }

	SDL_UnlockSurface(SDLVideo);
	//SDL_UpdateRect(SDLVideo,0,0,width-1,height-1);
	SDL_UpdateRect(SDLVideo,x1,y1,x2-x1+1,y2-y1+1);
	//SDL_UpdateRect(SDLVideo,0,height-20,width-1, height-1);
	}

static int
FrameChannel(char *fname)
	{
	char *ptr = fname + strlen(fname)-1;
	int ch = atoi(ptr-4);

	if (ch<0 || ch>6) ch = 0;

	return ch;
	}

void
display_image(struct Image *img)
	{
	int npixels;
	SDL_Event ev;
	static SDL_Overlay *so = NULL;
	static int Width=-1,Height=-1;
	static unsigned long tm;
	struct tm *t;
	char str[128];

	if (DetectBrokenFrames && isBrokenFrame(img))
	   return;

	if (FCount==0) tm = time(0);

	if (DrawFrame) {
	   if (GainComp) ApplyGainCompensation(img);

	   if (VideoCut) {
	      CalculateCutout(img);
	      CutOut(img);
	      }

	   npixels = img->width * img->height;
	   if (Width != img->width || Height != img->height) {
	      if (so) SDL_Quit();
	      SDL_BPP = 16;
	      so = init_display(img->width,img->height,SDL_BPP);
	      if (! so) {
		   SDL_BPP = 32;
	           so = init_display(img->width,img->height,SDL_BPP);
		   if (! so) FatalError("Cannot start SDL");
		   }

	      Width = img->width; Height = img->height;
	      }

#ifdef MSWIN32
	   sprintf(str,"%-20s %9u / %I64u",
		img->src_fname, img->tm_sec, img->tm_tsc);
#else
	   sprintf(str,"%-20s %9u / %llu",
		img->src_fname, img->tm_sec, img->tm_tsc);
#endif

	   DrawString(img, img->height - 20, 10, str);
	   SDLDisplay(img);
	   }
	else {
	   int c = FrameChannel(img->src_fname);
	   if (c == SeekColour) DrawFrame=1;
	   }

	if (Mode == MODE_STEP) Mode = MODE_PAUSE;

	while(1) {
	   while(SDL_PollEvent(&ev)) {
	      switch(ev.type) {
		case SDL_KEYDOWN: VKeyPress(ev.key.keysym.sym, ev.key.keysym.mod);
				  break;
		case SDL_KEYUP: VKeyRelease(ev.key.keysym.sym, ev.key.keysym.mod);
				  break;
		}
	     }
	   if (WriteFrame == 1) {
		printf("Write [%s]\n",img->dst_fname);
		WriteImage(img);
		WriteFrame=0;
		}

	   if (Mode == MODE_PLAY || Mode == MODE_STEP || Mode == MODE_NEXT) break;
	   Usleep(100000);
	   }

	++FCount;
	if (time(0) - tm > 2) {
	   if (img->tm_sec>0)
	      t = gmtime((const time_t *)(&img->tm_sec));
	   else t = NULL;

	   FCount /= 2;
	   printf("%-15.15s fps: %5.5d ",img->src_fname,FCount);
	   if (t != NULL)
		 printf("%02d:%02d:%02d", t->tm_hour, t->tm_min, t->tm_sec);
	   else  printf("00:00:00");
#ifdef MSWIN32
	   printf(" / %I64u\r", img->tm_tsc);
#else
	   printf(" / %llu\r", img->tm_tsc);
#endif
	   fflush(stdout);
	   FCount=0;
	   }

	return;
	}

static int
comp(const void *ptr1, const void *ptr2)
        {
        char **s1 = (char **)ptr1;
        char **s2 = (char **)ptr2;
        char *p1 = *s1;
        char *p2 = *s2;
        int i1,i2;

        // seek to end of string
        p1 += strlen(p1);
        p2 += strlen(p2);

        // seek back until we find the last path entry
        while(p1 != *s1 && *(p1-1) != '/' && *(p1-1) != '\\') --p1;
        while(p2 != *s2 && *(p2-1) != '/' && *(p2-1) != '\\') --p2;

        // Go forward until we find a digit
        while(*p1 && ! isdigit(*p1)) ++p1;
        while(*p2 && ! isdigit(*p2)) ++p2;

        // Hopefully we are now on a substring that can be used for comparison
        i1 = atoi(p1);
        i2 = atoi(p2);

        if (i1<i2) return -1;
        if (i1>i2) return 1;
        return 0;
        }

static int
Convert_8_16_bpp(struct Image *img)
	{
	int depth = img->depth;
	int width = img->width;
	int height = img->height;

	if (depth == 8) {
	   register unsigned int X;
	   int o = width * height;
	   unsigned short *buffer = Malloc(o * 2);
	   unsigned char *ptr = (unsigned char *)img->data;

	   while(o-- >= 0) {
		X = ptr[o]; X<<=8; buffer[o]=X;
		}

	   free(img->data);
	   img->data = (unsigned char *)buffer;

	   img->depth = 16;

	   if (img->type == IMG_FIT) {
		struct fits_info *fi = img->info;
		fi->bzero = 32768;
		}
	   }

	return 1;
	}

static char *filelist[10000];
static int FileCount=0;

void
playdir(char *dir)
	{
	struct dirent *e;
	int i;
	DIR *d = opendir(dir);
	char fname[256];

	Mode = MODE_PLAY;
	FileCount = 0;

	while(e = readdir(d)) {
	   if (strlen(e->d_name)>4 && e->d_name[0] != '.') {
		sprintf(fname,"%s/%s",dir,e->d_name);
		if (! access(fname,R_OK|F_OK) && isSupportedFile(fname)) {
		   filelist[FileCount++] = strdup(fname);
		   }
		}
	   }

	closedir(d);

	qsort(filelist,FileCount,sizeof(char *),comp);
	for(i=0; i<FileCount; ++i) {
	   if (Mode != MODE_NEXT) playfile(filelist[i]);
	   free(filelist[i]);
	   }

	FileCount=0;
	}

void
playfile(char *fname)
	{
	struct Image *img;
	int len = strlen(fname);
	static int count=0;

	if (! strcmp(fname+len-4,".fta")) 
	   playarchive(fname);
	else {
	   //printf("%-5d\r",count); count++; fflush(stdout);
	   img = LoadImage(fname,NULL);
	   display_image(img);
	   DestroyImage(img);
	   }
	}

void
playarchive(char *arch)
	{
	struct Image *img;
#ifndef MSWIN32
	extern FILE *fopen64();
	FILE *in = fopen64(arch,"rb");
#else
	FILE *in = fopen(arch,"rb");
#endif
	int count,err;

	if (! in) {
	   FatalError("Cannot open archive");
	   }

	FrameCount=0;
	FrameOffsets[0] = 0;

	while(1) {
	   if (fseek(in,FrameOffsets[FrameCount],SEEK_SET)) perror("");

	   if (Mode == MODE_NEXT) break;

	   if (DrawFrame)
	      img = ArchiveLoadNextImage(in,&err,ARCHIVE_LOAD);
	   else
	      img = ArchiveLoadNextImage(in,&err,ARCHIVE_SKIP);
		
	   if (err==0) break; // end of archive
	   if (err<0) FatalError("Error while reading archive");
	   //printf("%-5d (%-5d)\r",FrameCount,FrameOffsets[FrameCount]); fflush(stdout);
	   FrameCount++;
	   FrameOffsets[FrameCount] = ftell(in);

	   // Upgrade image to 16bpp
	   if (DrawFrame && img->depth == 8)
		Convert_8_16_bpp(img);

	   if (DrawFrame && PopFilter) pop_filter(img);
	   display_image(img);
	   DestroyImage(img);
	   }
	}
	   
	
void
SDLDisplayArea(int width, int height, SDL_Overlay *so)
{
  unsigned char *pimage;
  register int x,y;

    for (y=0;y<height;y++) {
      pimage=so->pixels[0]+y*so->pitches[0];
      for (x=0;x<width;x++) {
        pimage[x*2]=(unsigned char)(255-pimage[x*2]);
      }
    }
}

void
ByteSwap(unsigned char *data, int nbytes)
	{
	int i;

	while(nbytes) {
	   unsigned char tmp = *data;
	   *data = *(data+1);
	   *(data+1) = tmp;
	   nbytes -= 2;
	   }
	}

#define SHARPEN_DIV 6
void
Sharpen(unsigned int *buffer, int width, int height)
	{
	int x,y,o;
	static unsigned int *bptr = NULL;
	static int bufsize = 0;
	int npixels = width * height;

	if (npixels != bufsize) {
	   if (bptr != NULL) free(bptr);
	   bptr = Malloc(npixels * 4);
	   bufsize = npixels;
	   }

	for(o=y=2; y<height-2; ++y)
	   for(x=2; x<width-2; ++x,++o) {
		int v = buffer[o] * 3;
		v -= buffer[o-2]/SHARPEN_DIV;
		v -= buffer[o+2]/SHARPEN_DIV;
		v -= buffer[o-width-width-2]/SHARPEN_DIV;
		v -= buffer[o-width-width]/SHARPEN_DIV;
		v -= buffer[o-width-width+2]/SHARPEN_DIV;
		v -= buffer[o+width-width-2]/SHARPEN_DIV;
		v -= buffer[o+width+width]/SHARPEN_DIV;
		v -= buffer[o+width+width+2]/SHARPEN_DIV;
		//bptr[o] = (unsigned int)((double)v/2.5);
		if (v>0) bptr[o] = v; else bptr[o]=0;
		}

	memcpy(buffer,bptr,npixels*4);
	}

void
UnsharpMask(unsigned int *buffer, int width, int height, double m)
	{
	static int Npixels = -1;
	static unsigned int *bptr = NULL;
	int npixels = width * height;

	if (npixels != Npixels) {
	   if (bptr != NULL) free(bptr);
	   bptr = Malloc(npixels * 4);
	   Npixels = npixels;
	   }
	
	memcpy(bptr, buffer, npixels*4);
	Smooth(bptr, width, height);
	SubtractScale(buffer,bptr,npixels,m);
	Smooth(buffer,width, height);
	}

void
Smooth(unsigned int *buffer, int width, int height)
	{
	int x,y,o;
	static unsigned int *bptr = NULL;
	static int bufsize = 0;
	int npixels = width * height;

	if (npixels != bufsize) {
	   if (bptr != NULL) free(bptr);
	   bptr = Malloc(npixels * 4);
	   bufsize = npixels;
	   }

	for(o=y=0; y<height; ++y)
	   for(x=0; x<width; ++x,++o) {
		int v = buffer[o];
		v += buffer[o-1];
		v += buffer[o+1];
		v += buffer[o-width-1];
		v += buffer[o-width];
		v += buffer[o-width+1];
		v += buffer[o+width-1];
		v += buffer[o+width];
		v += buffer[o+width+1];
		bptr[o] = v/9;
		}

	memcpy(buffer,bptr,npixels*4);
	}

void
AlphaTrimmedMean(unsigned int *buffer, int width, int height)
	{
	static unsigned int *bptr = NULL;
	static int Npixels = 0;
	int npixels = width * height;
	int x,y,o,min,max;

	if (npixels != Npixels) {
	   if (bptr != NULL) free(bptr);
	   bptr = Malloc(npixels * 4);
	   Npixels = npixels;
	   }

	for(o=y=1; y<height-1; ++y)
	   for(x=1; x<width-1; ++x,++o) {
		int n,v = buffer[o];
		min=v; max=v;

		n=buffer[o-1];       v+=n; if (n<min) min=n; if (n>max) max=n;
		n=buffer[o+1];       v+=n; if (n<min) min=n; if (n>max) max=n;
		n=buffer[o-width-1]; v+=n; if (n<min) min=n; if (n>max) max=n;
		n=buffer[o-width];   v+=n; if (n<min) min=n; if (n>max) max=n;
		n=buffer[o-width+1]; v+=n; if (n<min) min=n; if (n>max) max=n;
		n=buffer[o+width-1]; v+=n; if (n<min) min=n; if (n>max) max=n;
		n=buffer[o+width];   v+=n; if (n<min) min=n; if (n>max) max=n;
		n=buffer[o+width+1]; v+=n; if (n<min) min=n; if (n>max) max=n;

		v -= min+max;
		bptr[o] = v/7;
		}

	memcpy(buffer,bptr,npixels*4);
	}

// buf1 -= buf2 * m
void
SubtractScale(unsigned int *buf1, unsigned int *buf2, int npixels, double m)
	{
	while(npixels-- > 0) {
	   int v = *buf1; v -= *buf2 * m;
	   if (v<0) v=0;
	   *buf1 = v;
	   buf1++; buf2++;
	   }
	}

#ifdef MSWIN32
#include <windows.h>

extern int main(int argc, char* argv[]);

int WINAPI WinMain(HINSTANCE instance, HINSTANCE prev_instance, char* command_line, int show_command)
{
    main(__argc, __argv);
}
#endif
 

int
main(int argc, char **argv)
	{
	struct Image *img;
	int i;

	Quiet=1;
	InputHistoStretch = 0;

	for(i=0; i<MAX_CHANNEL; ++i) {
	   Tables[i].map = NULL;
	   Tables[i].gain = 1.0;
	   Tables[i].gamma = 1.0;
	   }

	for(i=1; i<argc; ++i) {
	   char *arg = argv[i];

	   if (!strncmp(argv[i],"-istretch",9)) {
		InputHistoStretch = 1;
		continue;
		}

	   if (!strncmp(argv[i],"-stack=",7)) {
		VideoStack = atoi(argv[i]+7);
		continue;
		}

	   if (!strncmp(argv[i],"-cut=",5)) {
		int x,y;

		if (sscanf(argv[i]+5,"%d,%d",&x,&y)==2) {
		   CutX = x;
		   CutY = y;
		   }
		else {
		   int cut = atoi(argv[i]+5);
		   CutX = CutY = newWidth = newHeight = cut;
		   }

		DoCutout=1;
		VideoCut = 1;
		continue;
		}

	   if (!strncmp(argv[i],"-popfilter=",11)) {
		PopFilter = atoi(argv[i]+11);
		if (PopFilter<1) PopFilter = 3;
		continue;
		}

	   if (!strncmp(argv[i],"-offset=",6)) {
		Offset = atoi(argv[i]+8);
		continue;
		}

	   if (!strncmp(argv[i],"-gain=",6)) {
		double g = atof(argv[i]+6);
		int j;

		for(j=0; j<MAX_CHANNEL; ++j) Tables[j].gain = g;

		continue;
		}

	   if (!strncmp(argv[i],"-gain",5) && isdigit(argv[i][5])) {
		int t = atoi(argv[i]+5);
		Tables[t].gain = atof(argv[i]+7);
		continue;
		}

	   if (!strncmp(argv[i],"-amean",6)) {
		AMean = 1;
		continue;
		}

	   if (!strncmp(argv[i],"-dbf",3)) {
		DetectBrokenFrames=1;
		DBF_TYPE=DBF_PLANET;
		continue;
		}

	   if (!strncmp(argv[i],"-umask=",7)) {
		UMask = atof(argv[i]+7);
		continue;
		}

	   if (!strncmp(argv[i],"-gamma=",7)) {
		int j;
		double g = atof(argv[i]+7);

		for(j=0; j<MAX_CHANNEL; ++j) Tables[j].gamma = g;

		continue;
		}

	   if (!strncmp(argv[i],"-gaincomp=",10)) {
		GainComp = strdup(argv[i]+10);
		if (! LoadGainCompRef(GainComp)) {
		   printf("gaincomp %s failed\n",GainComp);
		   exit(1);
		   }
		continue;
		}

	   if (isDirectory(arg)) {
		printf("playing dir %s\n",arg);
	      	playdir(arg);
		}
	   else if (isFile(arg)) {
		printf("playing file %s\n",arg);
	      	playfile(arg);
		}
	   }

	SDL_Quit();
	}
